package org.convey;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.awt.geom.*;

import org.jdom.*;

/**
 * @author Andy Ames
 * @version 0.1
 */
public class FittedCurveGraphic extends CubicCurveGraphic {

	private transient Vector points = new Vector();

	private transient boolean repaintedPolyline = true;

	public FittedCurveGraphic() {
	}

	public FittedCurveGraphic(GraphicPanel graphicPanel) {
		super(graphicPanel);
	}

	public FittedCurveGraphic(GraphicPanel graphicPanel, Point p) {
		super(graphicPanel);
		repaintedPolyline = false;
		points.add(p);
		updateCache();
		super.repaint();
	}

	protected CubicCurve2D createFittedCurve() {

		int N = this.points.size();
		if(N < 4)
			return null;

		double[] x_2_N = new double[N];
		double[] y_2_N = new double[N];
		double[] t_N = new double[N];
		double[][] F_N_4 = new double[N][4];
		double[][] FT_4_N = new double[4][N];
		for(int i = 0; i < N; i++) {
			t_N[i] = (double)i / (double)N;

			Point pt = (Point)this.points.elementAt(i);
			x_2_N[i] = (double)pt.x;
			y_2_N[i] = (double)pt.y;

			FT_4_N[0][i] = F_N_4[i][0] = 1.0f;
			FT_4_N[1][i] = F_N_4[i][1] = t_N[i];
			FT_4_N[2][i] = F_N_4[i][2] = Math.pow(t_N[i], 2);
			FT_4_N[3][i] = F_N_4[i][3] = Math.pow(t_N[i], 3);
		}

		double[][] FTxF_4_4 = new double[4][4];
		for(int i = 0; i < 4; i++)
			for(int j = 0; j < 4; j++)
				for(int k = 0; k < N; k++)
					FTxF_4_4[i][j] += FT_4_N[i][k] * F_N_4[k][j];

		double[] FTxx_4 = new double[4];
		double[] FTyy_4 = new double[4];
		for(int i = 0; i < 4; i++) {
			for(int k = 0; k < N; k++) {
				FTxx_4[i] += FT_4_N[i][k] * x_2_N[k];
				FTyy_4[i] += FT_4_N[i][k] * y_2_N[k];
			}
		}

		Matrix Ax = new Matrix(FTxF_4_4);
		Matrix Cx = Ax.lu().solve(new Matrix(FTxx_4, FTxx_4.length));

		Matrix Ay = new Matrix(FTxF_4_4);
		Matrix Cy = Ay.lu().solve(new Matrix(FTyy_4, FTyy_4.length));

		Matrix Q  = new Matrix(4, 2);
		for(int i = 0; i < 4; i++) {
			Q.set(i,0,Cx.get(i,0));
			Q.set(i,1,Cy.get(i,0));
		}

		double[][] vals = { { 0.0, 0.0, 0.0, 1.0},
												{ 0.0, 0.0, 3.0, -3.0 },
												{ 0.0, 3.0, -6.0, 3.0 },
												{ 1.0, -3.0, 3.0, -1.0 }
											};
		Matrix B = new Matrix(vals);
		Matrix G = B.lu().solve(Q);
		CubicCurve2D bestFitCurve = new CubicCurve2D.Float((float)G.get(0,0), (float)G.get(0,1),
										 (float)G.get(1,0), (float)G.get(1,1),
										 (float)G.get(2,0), (float)G.get(2,1),
										 (float)G.get(3,0), (float)G.get(3,1));

		return bestFitCurve;
	}

	/**
	 * Invoked when a mouse button has been released on a component. If the
	 * preceding mouseDragged method collected less than 4 points for the points
	 * vector, the points vector is set to null as there is not enough points to
	 * use the curve fitting algorithm. Otherwise a new cubic curve graphic is
	 * created by calling the createFittedCurve method.
	 */
	public void mouseReleased(MouseEvent e) {
		if (super.getState() == CREATING) {
			if(points.size() >= 4) {
				super.curve = createFittedCurve();
				super.setState(DORMANT);
			} else
				super.setState(DELETE);
		} else
			super.mouseReleased(e);
	}

	/**
	* Invoked when a mouse button is pressed on a component and then
	* dragged.  Mouse drag events will continue to be delivered to
	* the component where the first originated until the mouse button is
	* released (regardless of whether the mouse position is within the
	* bounds of the component).
	*/
	public void mouseDragged(MouseEvent e) {
		if(super.getState() == CREATING) {
			points.add(e.getPoint());
			updateCache();
			super.repaint();
		} else
			super.mouseDragged(e);
	}

	/**
	 * Method getStatus
	 *
	 * @return   a String
	 *
	 */
	public String getStatus() {
		return "";
	}

	/**
	 *
	 */
	public void updateCache() {
		if(super.getState() != CREATING) {
			super.updateCache();

			if(!repaintedPolyline) {
				// make dirty include polyline so it gets cleared (painted over)
				GeneralPath path = new GeneralPath();
				Point p = (Point)points.elementAt(0);
				path.moveTo(p.x, p.y);
				for(int i = 1; i < points.size(); i++) {
					p = (Point)points.elementAt(i);
					path.lineTo(p.x, p.y);
				}
				super.currDirty = path.getBounds();
				super.currDirty.grow(5, 5);
				this.repaintedPolyline = true;
			}
		} else if(this.points.size() > 0) {

			// If we are CREATING...

			GeneralPath path = new GeneralPath();
			Point p = (Point)points.elementAt(0);
			path.moveTo(p.x, p.y);
			for(int i = 1; i < points.size(); i++) {
				p = (Point)points.elementAt(i);
				path.lineTo(p.x, p.y);
			}
			super.shape = path;
			super.pathHotSpot = new BasicStroke(6).createStrokedShape(super.shape);
			super.currDirty = super.shape.getBounds();
			super.currDirty.grow(5, 5);

			super.bounds = super.shape.getBounds();
		}
	}
}

