package org.convey;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import javax.swing.event.*;
import java.io.*;
import org.jdom.*;
import java.lang.reflect.*;
import java.util.*;

/**
 * Graphic abstract class extended by all Graphic objects.
 * <h3>Graphic State
 * A Graphic object can be in any one of the following states: CREATING,
 * DORMANT, SELECTED, or DELETE.
 * <p>
 * A Graphic in the CREATING state is still in the process of responding to
 * mouse or key events in order to construct itself. While in this state, the
 * Graphic has not yet been added to the GraphicPanel's graphics Vector.
 * However, the Graphic will still receive mouse and key events. In fact, it
 * will be the only Graphic to receive events. Additionally, its paint() method
 * will be called frequently in response to the GraphicPanel's repaint events.
 * <p>
 * A Graphic in the DORMANT state has already been created, but is not selected.
 * Graphics in GraphicPanel's clipboardGraphics Vectors are currently always in
 * the DORMANT state.
 * <p>
 * A Graphic in the SELECTED state is in the selectedGraphics Vector of the
 * GraphicPanel.
 * <p>
 * The DELETE state indicates to the GraphicPanel that the Graphic should be
 * removed from or not be added to the GraphicPanel's graphics Vector.
 * An example use of this state is FittedCurveGraphic when less than four points
 * were generated during creation. When this happens, the fitted curve is
 * underdetermined and therefore cannot be created.
 * <p>
 * Repainting of a Graphic is implemented using two dirty Rectangle instances:
 * currDirty and prevDirty. Initially, these two refereces point to the same
 * instance. Then, when a Graphic is modified by one of its methods, such as
 * {@link fitToRect} and {@link translate}, a new instance of currDirty is
 * created.
 * <p>
 * Finally, when {@link repaint} is called, if currDirty and prevDirty do not
 * refer to the same instance, their union is sent to the GraphicPanel for
 * repainting and prevDirty is set to equal currDirty. Otherwise, nothing
 * happens.
 * <p>
 * A repaint may be forced upon a Graphic via {@link forceRepaint}. This is
 * useful when a property of a Graphic changes that did not affect its geometry,
 * but did affect its display. The fill and stroke colors of ShapeGraphic are
 * examples of such properties.
 * <p>
 * Following is a list of justifications for the repaint design:
 * <ul>
 * <li>This simplifies code a bit because we keep creating oldDirty and unioning
 * it with new dirty every time something happens,
 * <li>It takes a HACK and turns it into the design,
 * <li>It ensures that no dirty region falls through the cracks,
 * <li>Makes the responsibility of repainting solely that of the Graphic
 * objects,
 * <li>It's flexible enough for future changes to the Graphic object
 * implementations.
 * </ul>
 * Every Graphic object has a unique name identifier. This name is used when
 * working on a networked GraphicPanel. When a Graphic object changes, a change
 * event is sent across the network. The Graphic object that changed is
 * identified with the name property.
 * <p>
 * All Graphic object's data is relative to the same coordinate system, where
 * the origin is at the upper-left corner of the GraphicPanel and a unit length
 * is one device pixel. As a result, operations on the data are subject to the
 * accumulation of error.
 * <p>
 * To minimize this error, operations such as {@link translate} and
 * {@link scale} that are performed in response to {@link mouseDragged} events
 * are computed using the data from the beginning of the current operation.
 * <p>
 * This data is named with "orig" as the prefix. The Graphic class defines
 * origBounds to be the bounding rectangle from before the current
 * {@link fitToRect} operation. All invocations of {@link fitToRect} will be
 * computed using this Rectangle and the new Rectangle.
 * <p>
 * The {@link setOrig} method is used to notify a Graphic that it should
 * initialize its origBounds property for an operation to begin.
 * <p>
 * Subclasses of Graphic will use a similar scheme. For example,
 * PolylineGraphic has a Vector property called origPoints that contains the
 * polyline points from before the operation began.
 *
 * @author Andy Ames
 * @version 0.3
 */
public abstract class Graphic
		implements MouseInputListener, KeyListener, Serializable {

	public static final int CREATING = 0;
	public static final int DORMANT = 1;
	public static final int SELECTED = 2;
	public static final int DELETE = 3;
	private static final int MAX_STATE = 3;

	public static final int TOP_LEFT = 0;
	public static final int TOP = 1;
	public static final int TOP_RIGHT = 2;
	public static final int RIGHT = 3;
	public static final int BOTTOM_RIGHT = 4;
	public static final int BOTTOM = 5;
	public static final int BOTTOM_LEFT = 6;
	public static final int LEFT = 7;

	private transient int state;

	protected static Hashtable colorToStringTable = new Hashtable();

	protected static Hashtable stringToColorTable = new Hashtable();

	static {
		colorToStringTable.put(new Color(0, 255, 255), "aqua");
		colorToStringTable.put(new Color(0, 0, 0), "black");
		colorToStringTable.put(new Color(0, 0, 255), "blue");
		colorToStringTable.put(new Color(255, 0, 255), "fuchsia");
		colorToStringTable.put(new Color(128, 128, 128), "gray");
		colorToStringTable.put(new Color(0, 255, 0), "green");
		colorToStringTable.put(new Color(0, 255, 0), "lime");
		colorToStringTable.put(new Color(128, 0, 0), "maroon");
		colorToStringTable.put(new Color(0, 0, 128), "navy");
		colorToStringTable.put(new Color(128, 128, 0), "olive");
		colorToStringTable.put(new Color(128, 0, 128), "purple");
		colorToStringTable.put(new Color(255, 0, 0), "red");
		colorToStringTable.put(new Color(192, 192, 192), "silver");
		colorToStringTable.put(new Color(0, 128, 128), "teal");
		colorToStringTable.put(new Color(255, 255, 255), "white");
		colorToStringTable.put(new Color(255, 255, 0), "yellow");

		stringToColorTable.put("aqua", new Color(0, 255, 255));
		stringToColorTable.put("black", new Color(0, 0, 0));
		stringToColorTable.put("blue", new Color(0, 0, 255));
		stringToColorTable.put("fuchsia", new Color(255, 0, 255));
		stringToColorTable.put("gray", new Color(128, 128, 128));
		stringToColorTable.put("grey", new Color(128, 128, 128));
		stringToColorTable.put("green", new Color(0, 255, 0));
		stringToColorTable.put("lime", new Color(0, 255, 0));
		stringToColorTable.put("maroon", new Color(128, 0, 0));
		stringToColorTable.put("navy", new Color(0, 0, 128));
		stringToColorTable.put("olive", new Color(128, 128, 0));
		stringToColorTable.put("purple", new Color(128, 0, 128));
		stringToColorTable.put("red", new Color(255, 0, 0));
		stringToColorTable.put("silver", new Color(192, 192, 192));
		stringToColorTable.put("teal", new Color(0, 128, 128));
		stringToColorTable.put("white", new Color(255, 255, 255));
		stringToColorTable.put("yellow", new Color(255, 255, 0));
	}

	protected String name;

	/**
	 *  Actual point pressed used for translation of graphic objects.
	 */
	protected transient Point pressPoint;

	/**
	 * Point information used for translation of graphic objects.
	 */
	protected transient Point lastDragPoint;

	/**
	 * Containing GraphicPanel instance.
	 */
	protected transient GraphicPanel graphicPanel;

	/**
	 * Tight rectangular bounds for any graphic object.
	 */
	protected transient Rectangle bounds;

	protected transient Rectangle currDirty = new Rectangle();

	protected transient Rectangle prevDirty = new Rectangle();

	/**
	 * Invisible version of the graphic shape for hit-detection with thick stroke.
	 * Large stroke increases hit sensitivity for selecting graphics with very
	 * thin stroke widths.
	 */
	protected transient Shape pathHotSpot;

	/**
	 * Scaling operation anchor point. This is set in {@link mousePressed} if the
	 * GraphicPanel state is SCALE and one of the eight scale anchor points has
	 * been pressed.
	 * <p>
	 * This is one of TOP_LEFT, TOP, TOP_RIGHT, RIGHT, BOTTOM_RIGHT, BOTTOM,
	 * BOTTOM_LEFT, and LEFT.
	 * <p>
	 * A value of -1 indicates none of the control handles was pressed. Its value
	 * is returned to -1 upon {@link mouseReleased}.
	 */
	protected transient int scaleAnchor;

	protected transient Rectangle origBounds;

	protected transient Rectangle lastFitToRect;

	/**
	 * Default constructor necessary for creating Graphics from a message
	 * received.
	 */
	public Graphic() {
		this.state = DORMANT;
		this.name = "graphic1";
		scaleAnchor = -1;
	}

	/**
	 * Constructor taking the containing GraphicPanel instance. Used when the
	 * local user is constructing a Graphic.
	 */
	public Graphic(GraphicPanel graphicPanel) {
		this.graphicPanel = graphicPanel;
		this.state = CREATING;
		this.name = "graphic1";
		scaleAnchor = -1;
	}

	/**
	 * @todo Create an exception class for invalid names. The message of the
	 * exception can then be used to communicate the reason for the error.
	 */
	public void setName(String name) {
		if(name != null && !name.equals("") && !this.graphicPanel.containsName(name)) {
			String oldName = this.name;
			this.name = name;
			if(!this.graphicPanel.isIncoming())
				this.graphicPanel.addGraphicEvent(new GraphicNameEvent(this, oldName, name));
		}
	}

	public String getName() {
		return this.name;
	}

	/**
	 * Paints the Graphic object onto the Graphics2D object.
	 */
	public abstract void paint(Graphics2D g2d);

	public void repaint() {
		if(this.currDirty != this.prevDirty) {
			this.graphicPanel.repaint(this.currDirty.union(this.prevDirty));
			this.prevDirty = this.currDirty;
		}
	}

	public void forceRepaint() {
		if(this.currDirty != this.prevDirty) {
			this.graphicPanel.repaint(this.currDirty.union(this.prevDirty));
			this.prevDirty = this.currDirty;
		} else
			this.graphicPanel.repaint(this.currDirty);
	}

	public void paintSelection(Graphics2D g2d) {
		switch(this.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				g2d.setStroke(new BasicStroke(1));

				g2d.setColor(Color.gray);

				g2d.draw(this.bounds);

				g2d.setColor(Color.black);

				g2d.fillRect(this.bounds.x - 2, this.bounds.y - 2, 5, 5);
				g2d.fillRect(this.bounds.x + this.bounds.width - 2, this.bounds.y - 2, 5, 5);
				g2d.fillRect(this.bounds.x + this.bounds.width - 2, this.bounds.y + this.bounds.height - 2, 5, 5);
				g2d.fillRect(this.bounds.x - 2 , this.bounds.y + this.bounds.height - 2, 5, 5);

				g2d.fillRect(this.bounds.x + this.bounds.width / 2 - 2, this.bounds.y - 2, 5, 5);
				g2d.fillRect(this.bounds.x + this.bounds.width / 2 - 2, this.bounds.y + this.bounds.height - 2, 5, 5);
				g2d.fillRect(this.bounds.x - 2, this.bounds.y + this.bounds.height / 2 - 2, 5, 5);
				g2d.fillRect(this.bounds.x + this.bounds.width - 2, this.bounds.y + this.bounds.height / 2 - 2, 5, 5);
				break;
		}
	}

	public Rectangle getBounds() {
		return (Rectangle)this.bounds.clone();
	}

	public Rectangle getOrigBounds() {
		return this.origBounds;
	}

	public boolean hitPath(Point p) {
		return this.pathHotSpot.contains(p);
	}

	public boolean hitPath(Rectangle r) {
		return this.pathHotSpot.intersects(r);
	}

	public boolean hitVertex(Point p) {
		if(this.getState() == Graphic.CREATING)
			return false;
		switch(this.graphicPanel.getState()) {
			case GraphicPanel.SCALE:
				int xcoin = -1; // -1 is not coincident, 0 is left, 1 is right, 2 is center

				if(p.x >= this.bounds.x - 3 && p.x <= this.bounds.x + 3) {
					// left edge
					xcoin = 0;
				} else if(p.x >= this.bounds.x + this.bounds.width - 3 && p.x <= this.bounds.x + this.bounds.width + 3) {
					// right edge
					xcoin = 1;
				}  else if(p.x >= this.bounds.x + this.bounds.width / 2 - 3 && p.x <= this.bounds.x + this.bounds.width / 2 + 3) {
					// center
					xcoin = 2;
				}

				if(xcoin == -1)
					return false;

				if(p.y >= this.bounds.y - 3 && p.y <= this.bounds.y + 3) {
					// top edge
					return true;
				} else if(p.y >= this.bounds.y + this.bounds.height - 3 && p.y <= this.bounds.y + this.bounds.height + 3) {
					// bottom edge
					return true;
				}  else if(p.y >= this.bounds.y + this.bounds.height / 2 - 3 && p.y <= this.bounds.y + this.bounds.height / 2 + 3) {
					// middle
					return true;
				}

				break;
		}

		return false;
	}

	/**
	 * Return the graphicPanel instance associated with this graphic object.
	 */
	public GraphicPanel getGraphicPanel() {
		return this.graphicPanel;
	}

	public void setGraphicPanel(GraphicPanel graphicPanel) {
		this.graphicPanel = graphicPanel;
	}

	/**
	 * Sets the state of the graphic instance. The possible states are CREATING,
	 * DORMANT, SELECTED, VERTEX_MOVING, and DELETE. Note that if the state is
	 * changing to or from DORMANT, we must repaint to hide/show this graphic's
	 * selection.
	 */
	public void setState(int state) {
		if(state <= MAX_STATE && state >= 0 && this.state != state) {
			this.state = state;

			if(this.state == SELECTED)
				this.graphicPanel.addSelectedGraphic(this);
			else if(this.state == DORMANT)
				this.graphicPanel.removeSelectedGraphic(this);

			if(state != DELETE) {
				updateCache();
				this.repaint();
			}
		}
	}

	/**
	 * Returns the state of the graphic instance. The possible states are CREATING,
	 * DORMANT, SELECTED, and DELETE.
	 */
	public int getState() {
		return this.state;
	}

	/**
	 * implemented by each graphic; the bounds, hotspot, and visible shape is
	 * calculated.
	 */
	public abstract void updateCache();

	/**
	 * translate the hotspot, visible shape, and bounding rectangle associated with
	 * the graphic being dragged.
	 */
	public void translate(int x, int y) {
		PathIterator pathIter = this.pathHotSpot.getPathIterator(null);
		GeneralPath path = new GeneralPath();
		float[] coords = new float[6];
		while(!pathIter.isDone()) {
			switch(pathIter.currentSegment(coords)) {
				case PathIterator.SEG_CLOSE:
					path.closePath();
					break;
				case PathIterator.SEG_CUBICTO:
					coords[0] += x; coords[1] += y;
					coords[2] += x; coords[3] += y;
					coords[4] += x; coords[5] += y;
					path.curveTo(coords[0], coords[1], coords[2], coords[3], coords[4], coords[5]);
					break;
				case PathIterator.SEG_LINETO:
					coords[0] += x; coords[1] += y;
					path.lineTo(coords[0], coords[1]);
					break;
				case PathIterator.SEG_MOVETO:
					coords[0] += x; coords[1] += y;
					path.moveTo(coords[0], coords[1]);
					break;
				case PathIterator.SEG_QUADTO:
					coords[0] += x; coords[1] += y;
					coords[2] += x; coords[3] += y;
					path.quadTo(coords[0], coords[1], coords[2], coords[3]);
					break;
			}
			pathIter.next();
		}
		this.pathHotSpot = path;

		this.bounds.translate(x, y);

		// Must create new reference for currDirty per repaint() spec!
		this.currDirty = (Rectangle)this.currDirty.clone();
		this.currDirty.translate(x, y);

		this.repaint();
	}

	/**
	 * returns jdom XML Element for this graphic. This is used for serializing
	 * graphic objects for network messaging, clipboard operations, file saving,
	 * undo management.
	 */
	public Element toElement() {
		// we use a dummy name. The subclass will reassign the name.
		Element element = new Element("dummy");

		element.setAttribute("id", this.name);

		return element;
	}

	public void fromElement(Element element) {
		this.name = element.getAttributeValue("id");
	}

	public void setOrig() {
		this.origBounds = (Rectangle)this.bounds.clone();
	}

	public void fitToRect(Rectangle r) {
		if(this.origBounds.width == 0 || this.origBounds.height == 0)
			return;
		float tx1 = (float)-this.origBounds.x;
		float ty1 = (float)-this.origBounds.y;
		float sx = (float)r.width / (float)this.origBounds.width;
		float sy = (float)r.height / (float)this.origBounds.height;
		float tx2 = (float)r.x;
		float ty2 = (float)r.y;

		PathIterator pathIter = this.pathHotSpot.getPathIterator(null);
		GeneralPath path = new GeneralPath();
		float[] coords = new float[6];
		while(!pathIter.isDone()) {
			switch(pathIter.currentSegment(coords)) {
				case PathIterator.SEG_CLOSE:
					path.closePath();
					break;
				case PathIterator.SEG_CUBICTO:
					coords[0] = (coords[0] + tx1) * sx + tx2; coords[1] = (coords[1] + ty1) * sy + ty2;
					coords[2] = (coords[2] + tx1) * sx + tx2; coords[3] = (coords[3] + ty1) * sy + ty2;
					coords[4] = (coords[4] + tx1) * sx + tx2; coords[5] = (coords[5] + ty1) * sy + ty2;
					path.curveTo(coords[0], coords[1], coords[2], coords[3], coords[4], coords[5]);
					break;
				case PathIterator.SEG_LINETO:
					coords[0] = (coords[0] + tx1) * sx + tx2; coords[1] = (coords[1] + ty1) * sy + ty2;
					path.lineTo(coords[0], coords[1]);
					break;
				case PathIterator.SEG_MOVETO:
					coords[0] = (coords[0] + tx1) * sx + tx2; coords[1] = (coords[1] + ty1) * sy + ty2;
					path.moveTo(coords[0], coords[1]);
					break;
				case PathIterator.SEG_QUADTO:
					coords[0] = (coords[0] + tx1) * sx + tx2; coords[1] = (coords[1] + ty1) * sy + ty2;
					coords[2] = (coords[2] + tx1) * sx + tx2; coords[3] = (coords[3] + ty1) * sy + ty2;
					path.quadTo(coords[0], coords[1], coords[2], coords[3]);
					break;
			}
			pathIter.next();
		}
		this.pathHotSpot = path;

		this.bounds = (Rectangle)r.clone();

		// Must create new reference for currDirty per repaint() spec!
		this.currDirty = (Rectangle)this.bounds.clone();
		this.currDirty.grow(5, 5);

		this.repaint();
	}

	/**
	 * Invoked when a mouse button has been clicked on a component.
	 */
	public void mouseClicked(MouseEvent e) {}

	/**
	 * Invoked when a mouse button has been pressed on a component.
	 */
	public void mousePressed(MouseEvent e) {
		Point p = e.getPoint();
		if(this.graphicPanel.getState() == GraphicPanel.SCALE) {
			int xcoin = -1; // -1 is not coincident, 0 is left, 1 is right, 2 is center
			int ycoin = -1; // -1 is not coincident, 0 is top, 1 is bottom, 2 is middle

			if(p.x >= this.bounds.x - 3 && p.x <= this.bounds.x + 3) {
				// left edge
				xcoin = 0;
			} else if(p.x >= this.bounds.x + this.bounds.width - 3 && p.x <= this.bounds.x + this.bounds.width + 3) {
				// right edge
				xcoin = 1;
			}  else if(p.x >= this.bounds.x + this.bounds.width / 2 - 3 && p.x <= this.bounds.x + this.bounds.width / 2 + 3) {
				// center
				xcoin = 2;
			}

			if(p.y >= this.bounds.y - 3 && p.y <= this.bounds.y + 3) {
				// top edge
				ycoin = 0;
			} else if(p.y >= this.bounds.y + this.bounds.height - 3 && p.y <= this.bounds.y + this.bounds.height + 3) {
				// bottom edge
				ycoin = 1;
			}  else if(p.y >= this.bounds.y + this.bounds.height / 2 - 3 && p.y <= this.bounds.y + this.bounds.height / 2 + 3) {
				// middle
				ycoin = 2;
			}

			scaleAnchor = -1;
			if(xcoin == 0) {
				if(ycoin == 0) {
					this.scaleAnchor = BOTTOM_RIGHT;
				} else if(ycoin == 1) {
					this.scaleAnchor = TOP_RIGHT;
				} else if(ycoin == 2) {
					this.scaleAnchor = RIGHT;
				}
			} else if(xcoin == 1) {
				if(ycoin == 0) {
					this.scaleAnchor = BOTTOM_LEFT;
				} else if(ycoin == 1) {
					this.scaleAnchor = TOP_LEFT;
				} else if(ycoin == 2) {
					this.scaleAnchor = LEFT;
				}
			} else if(xcoin == 2) {
				if(ycoin == 0) {
					this.scaleAnchor = BOTTOM;
				} else if(ycoin == 1) {
					this.scaleAnchor = TOP;
				}
			}
		}

		lastDragPoint = pressPoint = e.getPoint();
	}

	/**
	* Invoked when a mouse button is pressed on a component and then
	* dragged.  Mouse drag events will continue to be delivered to
	* the component where the first originated until the mouse button is
	* released (regardless of whether the mouse position is within the
	* bounds of the component).
	*/
	public void mouseDragged(MouseEvent e) {
		Point p = e.getPoint();

		if(this.state == SELECTED) {
			switch(this.graphicPanel.getState()) {
				case GraphicPanel.SELECT:
					this.translate(p.x - lastDragPoint.x, p.y - lastDragPoint.y);
					lastDragPoint = p;
					break;
				case GraphicPanel.SCALE:
					if(scaleAnchor != -1) {
						this.lastFitToRect = (Rectangle)this.origBounds.clone();
						switch(scaleAnchor) {
							case TOP_LEFT:
								this.lastFitToRect.x = this.origBounds.x;
								this.lastFitToRect.width = p.x - this.origBounds.x;

								this.lastFitToRect.y = this.origBounds.y;
								this.lastFitToRect.height = p.y - this.origBounds.y;
								break;
							case TOP:
								this.lastFitToRect.y = this.origBounds.y;
								this.lastFitToRect.height = p.y - this.origBounds.y;
								break;
							case TOP_RIGHT:
								this.lastFitToRect.x = p.x;
								this.lastFitToRect.width = this.origBounds.x + this.origBounds.width - p.x;

								this.lastFitToRect.y = this.origBounds.y;
								this.lastFitToRect.height = p.y - this.origBounds.y;
								break;
							case RIGHT:
								this.lastFitToRect.x = p.x;
								this.lastFitToRect.width = this.origBounds.x + this.origBounds.width - p.x;
								break;
							case BOTTOM_RIGHT:
								this.lastFitToRect.x = p.x;
								this.lastFitToRect.width = this.origBounds.x + this.origBounds.width - p.x;

								this.lastFitToRect.y = p.y;
								this.lastFitToRect.height = this.origBounds.y + this.origBounds.height - p.y;
								break;
							case BOTTOM:
								this.lastFitToRect.y = p.y;
								this.lastFitToRect.height = this.origBounds.y + this.origBounds.height - p.y;
								break;
							case BOTTOM_LEFT:
								this.lastFitToRect.x = this.origBounds.x;
								this.lastFitToRect.width = p.x - this.origBounds.x;

								this.lastFitToRect.y = p.y;
								this.lastFitToRect.height = this.origBounds.y + this.origBounds.height - p.y;
								break;
							case LEFT:
								this.lastFitToRect.x = this.origBounds.x;
								this.lastFitToRect.width = p.x - this.origBounds.x;
								break;
						}
						this.fitToRect(this.lastFitToRect);
						this.updateCache();
					} else {
						this.translate(p.x - lastDragPoint.x, p.y - lastDragPoint.y);
						lastDragPoint = p;
					}
					break;
			}
		}

		this.repaint();
	}

	/**
	 * Invoked when a mouse button has been released on a component.
	 */
	public void mouseReleased(MouseEvent e) {
		if(this.state == SELECTED) {
			if(this.scaleAnchor == -1) {
				Point p = e.getPoint();
				int x = p.x - pressPoint.x;
				int y = p.y - pressPoint.y;
				if(x != 0 || y != 0)
					this.graphicPanel.pushGraphicEvent(new TranslateGraphicEvent(this, x, y));
			} else {
				if(this.lastFitToRect != null) {
					this.graphicPanel.pushGraphicEvent(new FitToRectGraphicEvent(this, this.origBounds, this.lastFitToRect));
					this.scaleAnchor = -1;
					this.lastFitToRect = null;
				}
			}
		}
	}

		/**
	 * Invoked when the mouse button has been moved on a component
	 * (with no buttons or mouse buttons pressed)
	 */
	public void mouseMoved(MouseEvent e) {}

		/**
	 * Invoked when the mouse exits a component.
	 */
	public void mouseExited(MouseEvent e) {}

		/**
		 * Invoked when the mouse enters a component.
	 */
	public void mouseEntered(MouseEvent e) {}

	/**
	 * Invoked when a key has been typed. This event occurs when a key press is
	 * followed by a key release.
	 */
	public void keyTyped(KeyEvent e) {}

	/**
	 * Invoked when a key has been pressed.
	 */
	public void keyPressed(KeyEvent e) {}

	/**
	 * Invoked when a key has been released.
	 */
	public void keyReleased(KeyEvent e) {}

	public void undoGraphicEvent(GraphicEvent event) {
		if(event == null || this != event.getGraphic())
			throw new IllegalArgumentException("Invalid event");

		if(event instanceof GraphicNameEvent) {
			GraphicNameEvent graphicNameEvent = (GraphicNameEvent)event;

			this.name = graphicNameEvent.getOrigName();
		} else if(event instanceof NewGraphicEvent) {
			NewGraphicEvent newGraphicEvent = (NewGraphicEvent)event;

			this.graphicPanel.removeSelectedGraphic(this);
			this.setState(Graphic.DORMANT);
			this.graphicPanel.removeGraphic(this);
			this.forceRepaint();
		} else if(event instanceof DeleteGraphicEvent) {
			DeleteGraphicEvent deleteGraphicEvent = (DeleteGraphicEvent)event;

			this.graphicPanel.addGraphic(this);
			this.graphicPanel.addSelectedGraphic(this);
			this.setState(Graphic.SELECTED);
		} else if(event instanceof TranslateGraphicEvent) {
			TranslateGraphicEvent translateGraphicEvent = (TranslateGraphicEvent)event;

			this.translate(-translateGraphicEvent.getX(), -translateGraphicEvent.getY());
		} else if(event instanceof FitToRectGraphicEvent) {
			FitToRectGraphicEvent fitToRectGraphicEvent = (FitToRectGraphicEvent)event;

			this.setOrig();
			this.fitToRect(fitToRectGraphicEvent.getOrigRect());
		}
	}

	public void redoGraphicEvent(GraphicEvent event) {
		if(event instanceof NewGraphicEvent) {
			NewGraphicEvent newGraphicEvent = (NewGraphicEvent)event;
			Graphic graphic = newGraphicEvent.getGraphic();

			this.graphicPanel.addGraphic(graphic);
			this.graphicPanel.addSelectedGraphic(graphic);
			graphic.setState(Graphic.SELECTED);
		} else if(event instanceof DeleteGraphicEvent) {
			DeleteGraphicEvent deleteGraphicEvent = (DeleteGraphicEvent)event;
			Graphic graphic = deleteGraphicEvent.getGraphic();

			this.graphicPanel.removeSelectedGraphic(graphic);
			graphic.setState(Graphic.DORMANT);
			this.graphicPanel.removeGraphic(graphic);
			this.forceRepaint();
		} else if(event instanceof TranslateGraphicEvent) {
			TranslateGraphicEvent translateGraphicEvent = (TranslateGraphicEvent)event;
			Graphic graphic = translateGraphicEvent.getGraphic();

			graphic.translate(translateGraphicEvent.getX(), translateGraphicEvent.getY());
		} else if(event instanceof FitToRectGraphicEvent) {
			FitToRectGraphicEvent fitToRectGraphicEvent = (FitToRectGraphicEvent)event;
			Graphic graphic = fitToRectGraphicEvent.getGraphic();

			this.setOrig();
			graphic.fitToRect(fitToRectGraphicEvent.getNewRect());
		}
	}

	public static String colorToString(Color color) {
		if(color == null)
			return "none";
		else if(colorToStringTable.containsKey(color))
			return (String)colorToStringTable.get(color);
		else
			return colorToRGB(color);
	}

	public static String colorToRGB(Color color) {
		String red = Integer.toHexString(color.getRed()).toUpperCase();
		String green = Integer.toHexString(color.getGreen()).toUpperCase();
		String blue = Integer.toHexString(color.getBlue()).toUpperCase();

		if(red.length() == 1)
			red = "0" + red;
		if(green.length() == 1)
			green = "0" + green;
		if(blue.length() == 1)
			blue = "0" + blue;

		StringBuffer buffer = new StringBuffer("#");
		if(red.charAt(0) == red.charAt(1) &&
			 green.charAt(0) == green.charAt(1) &&
			 blue.charAt(0) == blue.charAt(1)) {
			buffer.append(red.charAt(0));
			buffer.append(green.charAt(0));
			buffer.append(blue.charAt(0));
		} else {
			buffer.append(red);
			buffer.append(green);
			buffer.append(blue);
		}

		return buffer.toString();
	}

	public static Color stringToColor(String str) {
		if(str.equals("none"))
			return null;
		else if(stringToColorTable.containsKey(str))
			return (Color)stringToColorTable.get(str);
		else
			return rgbToColor(str);
	}

	public static Color rgbToColor(String rgb) {
		if(rgb.length() == 4) {
			int red = Character.digit(rgb.charAt(1), 16);
			int green = Character.digit(rgb.charAt(2), 16);
			int blue = Character.digit(rgb.charAt(3), 16);
			return new Color((red << 4) + red,
											 (green << 4) + green,
											 (blue << 4) + blue);
		} else if(rgb.length() == 7) {
			int red0 = Character.digit(rgb.charAt(2), 16);
			int green0 = Character.digit(rgb.charAt(4), 16);
			int blue0 = Character.digit(rgb.charAt(6), 16);
			int red1 = Character.digit(rgb.charAt(1), 16);
			int green1 = Character.digit(rgb.charAt(3), 16);
			int blue1 = Character.digit(rgb.charAt(5), 16);
			return new Color((red1 << 4) + red0,
											 (green1 << 4) + green0,
											 (blue1 << 4) + blue0);
		}

		return null;
	}

}