package org.convey;

import java.util.*;

import java.awt.*;
import java.awt.geom.*;

import org.jdom.*;

/**
 * ShapeGraphic is a Graphic implementation for constructing and manipulating
 * arbitrary shapes.
 *
 * @author Andy Ames
 */
public abstract class ShapeGraphic extends Graphic {

	protected Color fillColor;

	protected Color strokeColor;

	protected Shape shape;

	public ShapeGraphic() {
		this.name = "shape1";
	}

	public ShapeGraphic(GraphicPanel graphicPanel) {
		super(graphicPanel);
		this.shape = null;
		this.fillColor = this.graphicPanel.getCanvasPanel().getConveyFrame().getFillColor();
		this.strokeColor = this.graphicPanel.getCanvasPanel().getConveyFrame().getStrokeColor();
		this.name = "shape1";
	}

	/**
	 * Returns JDOM XML Element for this graphic. This is used for serializing
	 * graphic objects for network messaging, clipboard operations, file saving,
	 * undo management.
	 */
	public Element toElement() {
		Element element = super.toElement();

		element.setAttribute("style", this.toStyle());

		return element;
	}

	public void fromElement(Element element) {
		super.fromElement(element);

		String style = element.getAttributeValue("style");
		StringTokenizer tokenizer = new StringTokenizer(style, ":;");

		while(tokenizer.hasMoreTokens()) {
			String name = tokenizer.nextToken();
			String value = tokenizer.nextToken();

			if(name.equals("fill"))
				this.fillColor = stringToColor(value);
			else if(name.equals("stroke"))
				this.strokeColor = stringToColor(value);
		}
	}

	public void paint(Graphics2D g2d) {
		if(this.shape != null) {
			if(this.fillColor != null) {
				g2d.setColor(this.fillColor);
				g2d.fill(this.shape);
			}

			if(this.strokeColor != null) {
				g2d.setStroke(new BasicStroke(1.5f));
				g2d.setColor(this.strokeColor);
				g2d.setStroke(new BasicStroke(1.5f));
				g2d.draw(this.shape);
			}
		}
	}

	public void setShape(Shape shape) {
		this.shape = shape;
		this.updateCache();
		this.repaint();
	}

	public Shape getShape() {
		return this.shape;
	}

	public void setFillColor(Color fill) {
		Color origColor = this.fillColor;
		this.fillColor = fill;
		if(!super.graphicPanel.isIncoming()) {
			ShapeGraphicEvent event = new ShapeGraphicEvent(this, ShapeGraphicEvent.TYPE_FILL,
					origColor, fill);
			this.graphicPanel.addGraphicEvent(event);
		}
		this.updateCache();
		super.repaint();
	}

	public Color getFillColor() {
		return this.fillColor;
	}

	public void setStrokeColor(Color stroke) {
		Color origColor = this.strokeColor;
		this.strokeColor = stroke;
		if(!super.graphicPanel.isIncoming()) {
			ShapeGraphicEvent event = new ShapeGraphicEvent(this, ShapeGraphicEvent.TYPE_STROKE,
					origColor, stroke);
			this.graphicPanel.addGraphicEvent(event);
		}
		super.forceRepaint();
	}

	public Color getStrokeColor() {
		return this.strokeColor;
	}

	public void fitToRect(Rectangle r) {
		if(this.origBounds.width == 0 || this.origBounds.height == 0)
			return;
		float tx1 = (float)-this.origBounds.x;
		float ty1 = (float)-this.origBounds.y;
		float sx = (float)r.width / (float)this.origBounds.width;
		float sy = (float)r.height / (float)this.origBounds.height;
		float tx2 = (float)r.x;
		float ty2 = (float)r.y;

		PathIterator pathIter = this.shape.getPathIterator(null);
		GeneralPath path = new GeneralPath();
		float[] coords = new float[6];
		while(!pathIter.isDone()) {
			switch(pathIter.currentSegment(coords)) {
				case PathIterator.SEG_CLOSE:
					path.closePath();
					break;
				case PathIterator.SEG_CUBICTO:
					coords[0] = (coords[0] + tx1) * sx + tx2; coords[1] = (coords[1] + ty1) * sy + ty2;
					coords[2] = (coords[2] + tx1) * sx + tx2; coords[3] = (coords[3] + ty1) * sy + ty2;
					coords[4] = (coords[4] + tx1) * sx + tx2; coords[5] = (coords[5] + ty1) * sy + ty2;
					path.curveTo(coords[0], coords[1], coords[2], coords[3], coords[4], coords[5]);
					break;
				case PathIterator.SEG_LINETO:
					coords[0] = (coords[0] + tx1) * sx + tx2; coords[1] = (coords[1] + ty1) * sy + ty2;
					path.lineTo(coords[0], coords[1]);
					break;
				case PathIterator.SEG_MOVETO:
					coords[0] = (coords[0] + tx1) * sx + tx2; coords[1] = (coords[1] + ty1) * sy + ty2;
					path.moveTo(coords[0], coords[1]);
					break;
				case PathIterator.SEG_QUADTO:
					coords[0] = (coords[0] + tx1) * sx + tx2; coords[1] = (coords[1] + ty1) * sy + ty2;
					coords[2] = (coords[2] + tx1) * sx + tx2; coords[3] = (coords[3] + ty1) * sy + ty2;
					path.quadTo(coords[0], coords[1], coords[2], coords[3]);
					break;
			}
			pathIter.next();
		}
		this.shape = path;

		super.fitToRect(r);
	}

	public void translate(int x, int y) {
		super.translate(x, y);

		PathIterator pathIter = this.shape.getPathIterator(null);
		GeneralPath path = new GeneralPath();
		float[] coords = new float[6];
		while(!pathIter.isDone()) {
			switch(pathIter.currentSegment(coords)) {
				case PathIterator.SEG_CLOSE:
					path.closePath();
					break;
				case PathIterator.SEG_CUBICTO:
					coords[0] += x; coords[1] += y;
					coords[2] += x; coords[3] += y;
					coords[4] += x; coords[5] += y;
					path.curveTo(coords[0], coords[1], coords[2], coords[3], coords[4], coords[5]);
					break;
				case PathIterator.SEG_LINETO:
					coords[0] += x; coords[1] += y;
					path.lineTo(coords[0], coords[1]);
					break;
				case PathIterator.SEG_MOVETO:
					coords[0] += x; coords[1] += y;
					path.moveTo(coords[0], coords[1]);
					break;
				case PathIterator.SEG_QUADTO:
					coords[0] += x; coords[1] += y;
					coords[2] += x; coords[3] += y;
					path.quadTo(coords[0], coords[1], coords[2], coords[3]);
					break;
			}
			pathIter.next();
		}
		this.shape = path;

		super.repaint();
	}

	public String toFillStyle() {
		return "fill:" + colorToString(this.fillColor);
	}

	public String toStrokeStyle() {
		return "stroke:" + colorToString(this.strokeColor);
	}

	public String toStyle() {
		StringBuffer buffer = new StringBuffer(this.toFillStyle());
		buffer.append(";");
		buffer.append(this.toStrokeStyle());
		return buffer.toString();
	}

}